
##############################################################################
#
# Copyright (c) 2005-2006 Balazs Ree <ree@greenfinity.hu> All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

'''\
The data logger
'''

from threading import RLock
from actions.base import Action
from actions import CommentAction
from logrecord import LogRecord

class DataLogger:
    '''Data logger
    
    The publisher logs data into this logger which is going to
    be retrieved at test generation.

    Data can be logged by the log method of the datalogger, this will
    vreate a new log record. Later additional
    data can be logged into the same logrecord by the log method
    of the record itself.

    Logging has now two phases:
    
    1. logging of request data (DataLogger.log)
    2. logging of response data (lastrecord.log)

    '''
    
    def __init__(self):
        self.logging_on = False
        self.lock = RLock()
        self.request_data = []
    
    # --
    # Logging
    # --
            
    def start_logging(self):
        self.lock.acquire()
        try:
            self.request_data = []
            self.logging_on = True
        finally:
            self.lock.release()
        
    def log(self, **kw):
        self.lock.acquire()
        try:
            if self.logging_on:
                data_entry = LogRecord(**kw)
                self.request_data.append(data_entry)
                # also return the data entry
                return data_entry
            else:
                return None
        finally:
            self.lock.release()
        
    def stop_logging(self):
        self.lock.acquire()
        try:
            self.logging_on = False
        finally:
            self.lock.release()
     
    def log_comment(self, text='', header='User comment', can_append=True):
        'Log a comment text'
        self.lock.acquire()
        try:
            if can_append and self.request_data and isinstance(self.request_data[-1], CommentAction):
                # append to last comment, if any
                self.request_data[-1].append(text)
            else:
                # make a new comment
                self.request_data.append(CommentAction(text, header=header))
        finally:
            self.lock.release()
 
    # --
    # Retrieve
    # --

    def get_all_data(self):
        self.lock.acquire()
        try:
            return list(self.request_data)
        finally:
            self.lock.release()

    def __getitem__(self, index):
        self.lock.acquire()
        try:
            return self.request_data[index]
        finally:
            self.lock.release()

    def __iter__(self):
        i = 0
        while 1:
            self.lock.acquire()
            try:
                try:
                    item = self.request_data[i]
                except IndexError:
                    break
            finally:
                self.lock.release()
            yield item
            i += 1
    
    # --
    # Set referers
    # --
    
    def get_referer_index(self, index):
        '''Get the referer for a given object item
        
        There is a referer index, that points to the previous object that
        redirected to this object. If it does not exist, it will be the index of self.

        We need this in the first place because there are other requests with 30x 
        mangling with our important requests. (Most notably, 304-s where the 
        good requests seem to be 302-s, but I guess there could be others.)
        
        There could be different other strategies to handle this:

        1. ignore the 304 responses and consider only 302-s (and maybe others)

        2. use the HTTP_REFERER headers, this would be good, but sometimes
           the client might not send these.

        So this solution seems to be the most robust choice, as the Location
        must always be present with a redirect.

        The heuristics is that each request is referred by the one that is the
        closest submitted to it. The referring is recursive, i.e. always the
        genuine request is remembered as referer.

        XXX Additional issue:

        If the request does not have a response, then it has a publish exception.
        If this is 'Unauthorized', then we assume it "redirected" to the next
        published request. I hope they are sequential and no other request can
        come by.
        '''
        self.lock.acquire()
        try:
            array = self.request_data
            try:
                # use a stored value
                return array[index].referer_index
            except AttributeError:
                referred_from = None
                # If the previous request was an unauthorized, 
                # then I should immediately take this as a redirect of it.
                for i in range(index-1, -1, -1):
                    if isinstance(array[i], Action):
                        continue
                    if array[i].response_is_unauthorized:
                        # the previous one was a response-less one (internal redirect?)
                        referred_from = i
                    break
                # generate the value based on redirect urls
                if referred_from is None:
                    for i in range(index-1, -1, -1):
                        # skip direct actions
                        if isinstance(array[i], Action):
                            continue
                        if hasattr(array[i], 'redirection_index'):
                            # this is redirecting to a known place... go on
                            continue
                        if array[i].response_redirect_url == array[index].url:
                            # ok, found a referrer.
                            referred_from = i
                            break
                # now take value recursively
                if referred_from is not None:
                    referer = self.get_referer_index(referred_from)
                    # mark that the referer has been chained!
                    # this prohibits its reusal for higher order elements
                    array[referer].redirection_index = index
                else:
                    # Not referred
                    referer = index
                # set the referer
                array[index].referer_index = referer
                return referer
        finally:
            self.lock.release()
