
##############################################################################
#
# Copyright (c) 2005-2006 Balazs Ree <ree@greenfinity.hu> All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

'''\
================================
Mechanize autogeneration support
================================

You only need this module at generation time, at runtime you need the
Products.mechagen-runtime module and even only if you use extended browser
methods (recommended)

However, it is best to inherit from MechagenSupport in the way
described below.


Usage of the MechagenSupport class
---------------------------------

To make sure that the mechagen hook is properly
invoked from setUp, create a custom Functional
testcase mixin::

    from Products.mechagen.generator import MechagenSupport

    class MyFunctional(MechagenSupport, ZopeTestCase.Functional, 
            ZopeTestCase.ZopeTestCase):
    
        # this is not strictly necessary, just an example...
        default_user = ZopeTestCase.user_name
        auth_dict = {'__ac_name': default_user,
            '__ac_password': 'secret'}

        ... add other methods specific for functional tests

    ZopeTestCase.utils.setupSiteErrorLog()
    ZopeTestCase.utils.startZServer()

Having done this, you can inherit your test cases from
Functional.

When you run the test, you can use the MECHAGEN
environment variable. To invoke automatical recording, set MECHAGEN to
an empty string. (e.g. ``MECHAGEN= ...zopectl test``)


To invoke re-recording of an existing method, set MECHAGEN 
to ``method=existing_method_name``.
(e.g. ``MECHAGEN=method=test_method ... zopectl test``)

In any case, it is best if you limit the test run to a particular
class. The first class in which tests will be found will be
considered as the class of the generation.

Note that if you *override* setUp from one of your classes,
the hook will not run unless you *call* mechagen_master_hook
at the end of the method.

Usage of the recorder
---------------------

Run your test case. At least one method in the test case should
be allowed.

If the MECHAGEN environment var is defined,
the recorder will take control. Follow the instructions.
'''

from urllib import urlencode
import re, sys, random, os.path, traceback, os
from textwrap import dedent
from cStringIO import StringIO
from Testing import ZopeTestCase
from writer import Writer
from zpublisher_support import patch_zpublisher
from processor import request_processor
from actions import CommentAction

# --
# Test case support
# --

class Quit(Exception):
    'Quit recording, fallback to normal tests'
    
# Exiting errors must descent from KeyboardInterrupt.
class ParamError(KeyboardInterrupt):
    'Parameter parsing error'

class Params(object):
    '''Parameters
    '''

    allowed_parms = ('method', 'quiet')
    method = None
    quiet = False
    
    def __init__(self, **kw):
        for key in kw.iterkeys():
            if not key in self.allowed_parms:
                raise ParamError, 'Not an allowed parameter: "%s"' % (key, )
        self.__dict__.update(kw)

class MechagenSupport:
# XXX making it an object breaks all setup stuff... Why?
#class MechagenSupport(object):
    '''mechagen support class

    To make sure, that the mechagen hook is properly
    invoked from setUp, create a custom Functional
    testcase mixin.

    Having done this, you can inherit your test cases from
    Functional.
    '''
        
    def setUp(self):
        'This will call the hook'
        # XXX This is a bit fragile...
        # run the setup from ZopeTestCase. Well, if you overwrite this, you
        # must manually invoke mechagen_master_hook from the end of it.
        #
        # XXX This should be done by "super", but it does not work since if we inherit
        # from "object", something gets broken (some setup code would not run)
        mro = list(self.__class__.__mro__)
        index = mro.index(MechagenSupport)
        for cls in mro[index+1:]:
            try:
                setUp = cls.setUp
            except AttributeError:
                pass
            else:
                setUp(self)
                break
        #super(MechagenSupport, self).setUp()
        # now setup has been finished, incl. afterSetup.
        # Invoke the master hook.
        self.mechagen_master_hook()
    
    def parse_params(self, params):
        '''Parse the one-line parameters.

        syntax:

        delimiter: ','
        Between delimiter: 'parm' or 'parm=value' 
        '''
        d = {}
        for param in params.split(','):
            pos = param.find('=')
            if pos == -1:
                # boolean option
                param = param.strip().lower()
                if not param:
                    continue
                if param.startswith('no'):
                    param = param[2:]
                    if not param:
                        raise ParamError, 'Not a valid parameter: "no"'
                    d[param] = False
                else:
                    d[param] = True
            else:
                # key = value option
                key, value = param[:pos], param[pos+1:]
                key = key.strip().lower()
                value = value.strip()
                if not key:
                    raise ParamError, 'Parameter name must be specified (for value "%s")' % (value, )
                d[key] = value
        # process them
        params = Params(**d)
        return params
                
    def mechagen_env_var(self):
        return os.environ.get('MECHAGEN', None)

    def mechagen_master_hook(self):
        '''Master hook of mechagen

        this should be called *after* the afterSetUp
        has finished.
        '''
        if not hasattr(request_processor, '_mechagen_run'):
            # run only once!
            mechagen_env = self.mechagen_env_var()
            if mechagen_env is not None:
                # need to set flag beforehand or we go into recursion
                # (reset calls setup)
                request_processor._mechagen_run = True
                # parse params
                params = self.parse_params(mechagen_env)
                # call method generation
                self.mechagen_record(params)
        # These methods can raise KeyboardInterrupt to exit, or
        # return if they deliberately want
        # to run the normal tests
        # instead.
            
    def mechagen_record(self, params):
        '''Recording with given params.

        '''
        
        if params.method is not None:
            
            # Will execute the method and do a generation of it.
            
            gen_method = getattr(self, params.method)
            
            def end_method(success):
                # ask what to do
                while 1:
                    print dedent('''\

                        Now enter:
                        S to Save this test (default),
                        D to Discard this test.''')
                    response = raw_input('Enter choice (S/d) > ')
                    response = response.strip().lower()
                    if response == '':
                        response = 's'
                    if response in ('s', 'd'):
                        break
                    print
                    print 'Wrong response.'
                    print
                saving = response in ('s', )
                exiting = True
                return saving, exiting

            # set the method name

            params._methodname = 'test_AG_%s' % (params.method, )
            
        else:

            # Will start recording tests in interaction with the console.
            
            def gen_method():
                if hasattr(self, 'portal'):
                    # plone support present
                    url = self.portal.absolute_url()
                else:
                    url = self.folder.absolute_url()
                    ## alternative way
                    ##url = 'http://%s:%d/%s' %(ZopeTestCase.utils._Z2HOST, ZopeTestCase.utils._Z2PORT, ZopeTestCase.folder_name)
                if hasattr(self, 'auth_dict'):
                    auth_dict = self.auth_dict
                    credentials = {}
                else:
                    auth_dict = {}
                    credentials = {
                        'username': ZopeTestCase.user_name,
                        'password': ZopeTestCase.user_password,
                        }
                if credentials:
                    credentials_txt = 'and login as %(username)s:%(password)s' % credentials
                else:
                    credentials_txt = ''
                if auth_dict:
                    url_txt = '%s?%s' % (url, urlencode(auth_dict))
                else:
                    url_txt = url
                message = '''\
                    Recording requests ...

                    ******************************************************************
                    You can start recording requests now. For example, direct your 
                    browser to:
                        %s
                    %s

                    You can also start from a different URL in the portal or use
                    other user authentication depending on what your test sets up.

                    The first request should preferebly be a non-submit
                    request (i.e. enter a direct URL first)
                    ******************************************************************
                    Your typing will be included in the generated code as comments.
                    To finish the test, type "END" in a new line and press Enter.
                    The first line will also be used as a docstring to the method.
                    ''' % (url_txt, credentials_txt)
                # XXX TODO credentials are just printed, but not actually handled in generation
                print dedent(message)

                while 1:
                    line = raw_input('> ')
                    if line.strip() == 'END':
                        break
                    request_processor.log_comment(text=line)        
                print 'Logging stopped.'
            
            def end_method(success):
                # ask what to do
                while 1:
                    print dedent('''\

                        Now enter:
                            E to save this test and Exit (default),
                            S to Save this test and Start a new one,
                            D to Discard this test and start a new one,
                            Q to discard this test and Quit.''')
                    response = raw_input('Enter choice (E/s/d/q) > ')
                    response = response.strip().lower()
                    if response == '':
                        response = 'e'
                    if response in ('e', 's', 'd', 'q'):
                        break
                    print 'Wrong response.'
                    print
                saving = response in ('e', 's')
                exiting = response in ('e', 'q')
                return saving, exiting

            # Make method name autogenerated
            params._methodname = None

        # invoke the action itself
        params._gen_method_hook = gen_method
        params._end_method_hook = end_method
        params._save_method_hook = self._save_to_module_file
        params._err_method_hook = self._handle_error
        self._mechagen_mech(params)
        
    # --
    # Private parts
    # --
    
    def _mechagen_mech(self, params):
        
        # Iterate to record multiple tests.
        while 1:
            
            exiting = self.mechagen_single_session(params)
            
            if exiting:
                # Exiting
                if not params.quiet:
                    print 'Exiting from test generation.'
                    print
                # really exit. KeyboardInterrupt is the only way to pass through the
                # bastion of unittest. We arrange to print no error, though.
                sys.excepthook = lambda t, v, tb: None 
                raise KeyboardInterrupt
            else:
                # Restarting
                if not params.quiet:
                    print 'Starting new test.'
                # needs to reset test sandbox machinery explicitly
                # before running the next test
                request_processor.reset_app(self)

    def mechagen_single_session(self, params):
        '''Do a single generation session

        Returns True if wants to exit
        '''

        request_processor.start_logging()
            
        try:
            params._gen_method_hook()
        except Quit:
            # returning will just continue the program flow
            # and execute the tests as if without mechagen.
            return

        request_processor.stop_logging()
            
        if not params.quiet:
            print 'Generating code ...',
        
        # now replay.
        request_processor.replay(self)
        
        # count actions for info
        nr_actions = 0
        for action in request_processor.actions:
            # do not count comments, check actions
            if action.realaction:
                nr_actions += 1
            
        if not params.quiet:
            print ' for %s request actions ...' % nr_actions
        buf = StringIO()
        request_processor.generate(self, buf, methodonly=True)
        
        if not params.quiet:
            print 'Replaying generated code ...'
        (t, v, tb) = request_processor.check_generated(self, buf.getvalue())
        
        success = (v is None)
        if success:
            if not params.quiet:
                print 'Replay succeeded without errors.'
        else:
            params._err_method_hook(t, v, tb, params)
            
        saving, exiting = params._end_method_hook(success)
        
        if saving:
            # saving
            if not params.quiet:
                
                params._save_method_hook(params)

                print
                print '*' * 20
                print 'Test saved.'
                print '*' * 20
                print
        else:
            # Discarding
            if not params.quiet:
                print
                print '*' * 20
                print 'Test discarded.'
                print '*' * 20
                print

        # return exit status
        return exiting

    def _handle_error(self, t, v, tb, params):
        writer = Writer(StringIO())
        writer.write('''\
            Error during replay:'''
            )
        writer.newline()
        writer.newline()
        writer.indenton()
        writer.write(''.join(traceback.format_exception(t, v, tb, limit=20)))
        writer.indentoff()
        writer.newline()
        writer.write('''\
            This means that the generated code is faulty.
            Debug results for more information.'''
            )
        writer.newline()
        errmsg = writer.stream.getvalue()
        if not params.quiet:
            print "*" * 70
            print errmsg
            print "*" * 70
        request_processor.actions.append(CommentAction(errmsg, 'Error during replay'))

    def _save_to_module_file(self, params):

        # TODO make this real params
        fname = None # params.fname
        class_name = None # params.class_name
        methodname = params._methodname
        
        if fname is None:
            fname = sys.modules[self.__module__].__file__
            if fname.endswith('.pyc') or fname.endswith('.pyo'):
                fname = fname[:-1]

        if class_name is None:
            class_name = self.__class__.__name__

        if methodname is None:
            methodname = 'test_AG_%08d' % random.randint(0, 99999999)

        # read in the file
        origlines = file(fname).readlines()

        # make a backup

        f = file('%s/.%s.bup' % os.path.split(fname), 'w')
        for line in origlines:
            f.write(line)
        f.close()

        # all checks suppose 4-space indents
        re_classdef = re.compile(r'class +' + re.escape(class_name) + ' *\([^\)]*\) *:')
        re_classdef_end = re.compile(r'[^ \n\t#]')

        # write the modified file now.
        f = file(fname, 'w')
        def writegen():
            request_processor.generate(self, f, methodname=methodname)
            f.write('\n\n')
        where = 0
        for line in origlines:
            if where == 0:
                # before class
                if re_classdef.match(line):
                    where = 1
            elif where == 1:
                # inside class
                if re_classdef_end.match(line):
                    # This is the moment to insert the generated stuff
                    writegen()
                    where = 2
            f.write(line)
        if where < 2:
            # we put it to the end
            f.write('\n')
            f.write('# ' + '*' * 65 + '\n')
            f.write('# WARNING: Did not find insert position, inserting to end of file.\n')
            writegen()
        f.close()
        
    # Automatically patch the ZPublisher
    # when this class is imported
    patch_zpublisher()
